#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

import STPyV8


class TestContext(unittest.TestCase):
    def testEval(self):
        with STPyV8.JSContext() as context:
            self.assertEqual(2, context.eval("1+1"))
            self.assertEqual("Hello world", context.eval("'Hello ' + 'world'"))

    def testMultiNamespace(self):
        self.assertTrue(not bool(STPyV8.JSContext.inContext))
        self.assertTrue(not bool(STPyV8.JSContext.entered))

        class Global:
            name = "global"

        g = Global()

        with STPyV8.JSContext(g) as ctxt:
            self.assertTrue(ctxt)
            self.assertTrue(bool(STPyV8.JSContext.inContext))
            self.assertEqual(g.name, str(STPyV8.JSContext.entered.locals.name))

            class Local:
                name = "local"

            l = Local()

            with STPyV8.JSContext(l):
                self.assertTrue(bool(STPyV8.JSContext.inContext))
                self.assertEqual(l.name, str(STPyV8.JSContext.entered.locals.name))

            self.assertTrue(bool(STPyV8.JSContext.inContext))
            self.assertEqual(g.name, str(STPyV8.JSContext.current.locals.name))

        self.assertTrue(not bool(STPyV8.JSContext.entered))
        self.assertTrue(not bool(STPyV8.JSContext.inContext))

    def testMultiContext(self):
        with STPyV8.JSContext() as ctxt0:
            ctxt0.securityToken = "password"

            global0 = ctxt0.locals
            global0.custom = 1234

            self.assertEqual(1234, int(global0.custom))

            with STPyV8.JSContext() as ctxt1:
                ctxt1.securityToken = ctxt0.securityToken

                global1 = ctxt1.locals
                global1.custom = 1234

                with ctxt0:
                    self.assertEqual(1234, int(global0.custom))

                self.assertEqual(1234, int(global1.custom))

    def testPromiseResolved(self):
        with STPyV8.JSContext() as ctxt:
            ctxt.eval(
                """
                var message;
                let done = true;

                const isItDoneYet = new Promise((resolve, reject) => {
                    if (done) {
                        const workDone = 'Here is the thing I built'
                        resolve(workDone)
                    } else {
                        const why = 'Still working on something else'
                        reject(why)
                    }
                })

                const checkIfItsDone = () => {
                    isItDoneYet.then(ok => {
                        message = ok;
                    }).catch(err => {
                        message = err;
                    })
                }

                checkIfItsDone()
            """
            )

            self.assertEqual("Here is the thing I built", ctxt.locals.message)

    def testPromiseRejected(self):
        with STPyV8.JSContext() as ctxt:
            ctxt.eval(
                """
                var message;
                let done = false;

                const isItDoneYet = new Promise((resolve, reject) => {
                    if (done) {
                        const workDone = 'Here is the thing I built'
                        resolve(workDone)
                    } else {
                        const why = 'Still working on something else'
                        reject(why)
                    }
                })

                const checkIfItsDone = () => {
                    isItDoneYet.then(ok => {
                        message = ok;
                    }).catch(err => {
                        message = err;
                    })
                }

                checkIfItsDone()
            """
            )

            self.assertEqual("Still working on something else", ctxt.locals.message)

    def testSecurityChecks(self):
        with STPyV8.JSContext() as env1:
            env1.securityToken = "foo"

            # Create a function in env1.
            env1.eval("spy = function(){return spy;}")

            spy = env1.locals.spy

            self.assertTrue(isinstance(spy, STPyV8.JSFunction))

            # Create another function accessing global objects.
            env1.eval("spy2 = function(){return 123;}")

            spy2 = env1.locals.spy2

            self.assertTrue(isinstance(spy2, STPyV8.JSFunction))

            # Switch to env2 in the same domain and invoke spy on env2.
            env2 = STPyV8.JSContext()
            env2.securityToken = "foo"

            with env2:
                result = spy.apply(env2.locals)
                self.assertTrue(isinstance(result, STPyV8.JSFunction))

            env2.securityToken = "bar"

            # FIXME
            # Call cross_domain_call, it should throw an exception
            # with env2:
            #    self.assertRaises(STPyV8.JSError, spy2.apply, env2.locals)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import with_statement
from __future__ import print_function

import os
import sys
import re
import collections.abc

import _STPyV8

__version__ = _STPyV8.JSEngine.version

__all__ = [
    "ReadOnly",
    "DontEnum",
    "DontDelete",
    "Internal",
    "JSError",
    "JSObject",
    "JSNull",
    "JSUndefined",
    "JSArray",
    "JSFunction",
    "JSClass",
    "JSEngine",
    "JSContext",
    "JSIsolate",
    "JSStackTrace",
    "JSStackFrame",
    "JSScript",
    "JSLocker",
    "JSUnlocker",
    "JSPlatform",
]


# ICU
ICU_DATA_FOLDERS_UNIX = (
    "/usr/share/stpyv8",
    os.path.expanduser("~/.local/share/stpyv8"),
)
ICU_DATA_FOLDERS_OSX = (
    "/Library/Application Support/STPyV8",
    os.path.expanduser("~/Library/Application Support/STPyV8"),
)
ICU_DATA_FOLDERS_WINDOWS = (
    os.path.join(os.environ["PROGRAMDATA"], "STPyV8")
    if "PROGRAMDATA" in os.environ
    else None,
    os.path.join(os.environ["APPDATA"], "STPyV8") if "APPDATA" in os.environ else None,
)

icu_data_folders = None
if os.name in ("posix",):
    icu_data_folders = (
        ICU_DATA_FOLDERS_OSX if sys.platform in ("darwin",) else ICU_DATA_FOLDERS_UNIX
    )
else:
    icu_data_folders = ICU_DATA_FOLDERS_WINDOWS


class JSAttribute:
    def __init__(self, name):
        self.name = name

    def __call__(self, func):
        setattr(func, f"__{self.name}__", True)

        return func


ReadOnly = JSAttribute(name="readonly")
DontEnum = JSAttribute(name="dontenum")
DontDelete = JSAttribute(name="dontdel")
Internal = JSAttribute(name="internal")


class JSError(Exception):
    def __init__(self, impl):
        Exception.__init__(self)
        self._impl = impl

    def __str__(self):
        return str(self._impl)

    def __getattribute__(self, attr):
        impl = super().__getattribute__("_impl")

        try:
            return getattr(impl, attr)
        except AttributeError:
            return super().__getattribute__(attr)

    RE_FRAME = re.compile(
        r"\s+at\s(?:new\s)?(?P<func>.+)\s\((?P<file>[^:]+):?(?P<row>\d+)?:?(?P<col>\d+)?\)"
    )
    RE_FUNC = re.compile(r"\s+at\s(?:new\s)?(?P<func>.+)\s\((?P<file>[^\)]+)\)")
    RE_FILE = re.compile(r"\s+at\s(?P<file>[^:]+):?(?P<row>\d+)?:?(?P<col>\d+)?")

    @staticmethod
    def parse_stack(value):
        stack = []

        def int_or_nul(value):
            return int(value) if value else None

        for line in value.split("\n")[1:]:
            m = JSError.RE_FRAME.match(line)

            if m:
                stack.append(
                    (
                        m.group("func"),
                        m.group("file"),
                        int_or_nul(m.group("row")),
                        int_or_nul(m.group("col")),
                    )
                )
                continue

            m = JSError.RE_FUNC.match(line)

            if m:
                stack.append((m.group("func"), m.group("file"), None, None))
                continue

            m = JSError.RE_FILE.match(line)

            if m:
                stack.append(
                    (
                        None,
                        m.group("file"),
                        int_or_nul(m.group("row")),
                        int_or_nul(m.group("col")),
                    )
                )
                continue

            assert line

        return stack

    @property
    def frames(self):
        return self.parse_stack(self.stackTrace)


_STPyV8._JSError._jsclass = JSError  # pylint:disable=protected-access

JSObject = _STPyV8.JSObject
JSNull = _STPyV8.JSNull
JSUndefined = _STPyV8.JSUndefined
JSArray = _STPyV8.JSArray
JSFunction = _STPyV8.JSFunction
JSPlatform = _STPyV8.JSPlatform


class JSLocker(_STPyV8.JSLocker):
    def __enter__(self):
        self.enter()

        if JSContext.entered:
            self.leave()
            raise RuntimeError("Lock should be acquired before entering the context")

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if JSContext.entered:
            self.leave()
            raise RuntimeError("Lock should be released after leaving the context")

        self.leave()

    def __bool__(self):
        return self.entered()


class JSUnlocker(_STPyV8.JSUnlocker):
    def __enter__(self):
        self.enter()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.leave()

    def __bool__(self):
        return self.entered()


class JSClass:
    __properties__ = {}
    __watchpoints__ = {}

    def __getattr__(self, name):
        if name == "constructor":
            return JSClassConstructor(self.__class__)

        if name == "prototype":
            return JSClassPrototype(self.__class__)

        prop = self.__dict__.setdefault("__properties__", {}).get(name, None)

        if prop and isinstance(prop[0], collections.abc.Callable):
            return prop[0]()

        raise AttributeError(name)

    def __setattr__(self, name, value):
        prop = self.__dict__.setdefault("__properties__", {}).get(name, None)

        if prop and isinstance(prop[1], collections.abc.Callable):
            return prop[1](value)

        return object.__setattr__(self, name, value)

    def toString(self):
        """
        Return the string representation of the object
        """
        return f"[object {self.__class__.__name__}]"

    def toLocaleString(self):
        """
        Return the string representation of the object as a string value
        appropriate to the environment current locale
        """
        return self.toString()

    def valueOf(self):
        """
        Return the primitive value of the object
        """
        return self

    def hasOwnProperty(self, name):
        """
        Return a boolean value indicating whether the object has a property
        with the specified name
        """
        return hasattr(self, name)

    def isPrototypeOf(self, obj):
        """
        Return a boolean value indicating whether the object exists in the
        prototype chain of another object
        """
        raise NotImplementedError()

    def __defineGetter__(self, name, getter):
        """
        Bind the object property to a function to be called when that property
        is looked up
        """
        self.__properties__[name] = (getter, self.__lookupSetter__(name))

    def __lookupGetter__(self, name):
        """
        Return the function bound as a getter to the specified property
        """
        return self.__properties__.get(name, (None, None))[0]

    def __defineSetter__(self, name, setter):
        """
        Bind the object property to a function to be called when an attempt
        is made to set that property
        """
        self.__properties__[name] = (self.__lookupGetter__(name), setter)

    def __lookupSetter__(self, name):
        """
        Return the function bound as setter to the specified property
        """
        return self.__properties__.get(name, (None, None))[1]

    def watch(self, prop, handler):
        """
        Watch for a property to be assigned a value and runs a function when
        such assignment occurs
        """
        self.__watchpoints__[prop] = handler

    def unwatch(self, prop):
        """
        Remove a watchpoint set with the watch method
        """
        del self.__watchpoints__[prop]


class JSClassConstructor(JSClass):  # pylint:disable=abstract-method
    def __init__(self, cls):
        self.cls = cls

    @property
    def name(self):
        return self.cls.__name__

    def toString(self):
        return f"function {self.name}() {{\n  [native code]\n}}"

    def __call__(self, *args, **kwds):
        return self.cls(*args, **kwds)


class JSClassPrototype(JSClass):  # pylint:disable=abstract-method
    def __init__(self, cls):
        self.cls = cls

    @property
    def constructor(self):
        return JSClassConstructor(self.cls)

    @property
    def name(self):
        return self.cls.__name__


class JSEngine(_STPyV8.JSEngine):
    def __init__(self):
        _STPyV8.JSEngine.__init__(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        del self


JSScript = _STPyV8.JSScript
JSStackTrace = _STPyV8.JSStackTrace
JSStackTrace.Options = _STPyV8.JSStackTraceOptions
JSStackTrace.GetCurrentStackTrace = staticmethod(
    lambda frame_limit,  # pylint:disable=unnecessary-lambda
    options: _STPyV8.JSIsolate.current.GetCurrentStackTrace(frame_limit, options)
)
JSStackFrame = _STPyV8.JSStackFrame


class JSIsolate(_STPyV8.JSIsolate):
    def __enter__(self):
        self.enter()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.leave()
        del self


class JSContext(_STPyV8.JSContext):
    def __init__(self, obj=None, ctxt=None):
        self.lock = JSLocker()
        self.lock.enter()

        if ctxt:
            _STPyV8.JSContext.__init__(self, ctxt)
        else:
            _STPyV8.JSContext.__init__(self, obj)

    def __enter__(self):
        self.enter()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.leave()

        if hasattr(JSLocker, "lock"):
            self.lock.leave()
            self.lock = None

        del self


def icu_sync():
    if sys.version_info < (3, 10):
        from importlib_resources import files
    else:
        from importlib.resources import files

    for folder in icu_data_folders:
        if not folder or not os.path.exists(folder):
            continue

        version_file = os.path.join(folder, "stpyv8-version.txt")
        if not os.path.exists(version_file):
            continue

        with open(version_file, encoding="utf-8", mode="r") as fd:
            version = fd.read()

        if version.strip() in (__version__,):
            return

    try:
        stpyv8_icu_files = files("stpyv8-icu")
    except ModuleNotFoundError:
        return

    for f in stpyv8_icu_files.iterdir():
        if f.name not in ("icudtl.dat",):
            continue

        data = f.read_bytes()

        for folder in icu_data_folders:
            try:
                os.makedirs(folder, exist_ok=True)
                with open(os.path.join(folder, "icudtl.dat"), mode="wb") as fd:
                    fd.write(data)

                version_file = os.path.join(folder, "stpyv8-version.txt")
                with open(version_file, encoding="utf-8", mode="w") as fd:
                    fd.write(__version__)
            except PermissionError:
                pass


icu_sync()

v8_default_platform = JSPlatform()
v8_default_platform.init()

v8_default_isolate = JSIsolate()
v8_default_isolate.enter()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

import STPyV8


class TestIsolate(unittest.TestCase):
    def testBase(self):
        with STPyV8.JSIsolate() as isolate:
            self.assertIsNotNone(isolate.current)
            self.assertFalse(isolate.locked)

    def testEnterLeave(self):
        with STPyV8.JSIsolate() as isolate:
            self.assertIsNotNone(isolate.current)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
import datetime
import unittest
import pytest

import STPyV8

STPYV8_DEBUG = os.getenv("STPYV8_DEBUG") is not None


def convert(obj):
    if isinstance(obj, STPyV8.JSArray):
        return [convert(v) for v in obj]

    if isinstance(obj, STPyV8.JSObject):
        return dict([[str(k), convert(obj.__getattr__(str(k)))] for k in obj.__dir__()])

    return obj


class TestWrapper(unittest.TestCase):
    def testObject(self):
        with STPyV8.JSContext() as ctxt:
            o = ctxt.eval("new Object()")

            self.assertTrue(hash(o) > 0)

            o1 = o.clone()

            self.assertEqual(hash(o1), hash(o))
            self.assertTrue(o != o1)

        self.assertRaises(UnboundLocalError, o.clone)

    def testAutoConverter(self):
        with STPyV8.JSContext() as ctxt:
            ctxt.eval(
                """
                var_i = 1;
                var_f = 1.0;
                var_s = "test";
                var_b = true;
                var_s_obj = new String("test");
                var_b_obj = new Boolean(true);
                var_f_obj = new Number(1.5);
            """
            )

            _vars = ctxt.locals

            var_i = _vars.var_i

            self.assertTrue(var_i)
            self.assertEqual(1, int(var_i))

            var_f = _vars.var_f

            self.assertTrue(var_f)
            self.assertEqual(1.0, float(_vars.var_f))

            var_s = _vars.var_s
            self.assertTrue(var_s)
            self.assertEqual("test", str(_vars.var_s))

            var_b = _vars.var_b
            self.assertTrue(var_b)
            self.assertTrue(bool(var_b))

            self.assertEqual("test", _vars.var_s_obj)
            self.assertTrue(_vars.var_b_obj)
            self.assertEqual(1.5, _vars.var_f_obj)

            attrs = dir(ctxt.locals)

            self.assertTrue(attrs)
            self.assertTrue("var_i" in attrs)
            self.assertTrue("var_f" in attrs)
            self.assertTrue("var_s" in attrs)
            self.assertTrue("var_b" in attrs)
            self.assertTrue("var_s_obj" in attrs)
            self.assertTrue("var_b_obj" in attrs)
            self.assertTrue("var_f_obj" in attrs)

    def testExactConverter(self):
        class MyInteger(int, STPyV8.JSClass):
            pass

        class MyString(str, STPyV8.JSClass):
            pass

        class MyDateTime(datetime.time, STPyV8.JSClass):
            pass

        class Global(STPyV8.JSClass):
            var_bool = True
            var_int = 1
            var_float = 1.0
            var_str = "str"
            var_datetime = datetime.datetime.now()
            var_date = datetime.date.today()
            var_time = datetime.time()

            var_myint = MyInteger()
            var_mystr = MyString("mystr")
            var_mytime = MyDateTime()

        with STPyV8.JSContext(Global()) as ctxt:
            typename = ctxt.eval(
                "(function (name) { return this[name].constructor.name; })"
            )
            typeof = ctxt.eval("(function (name) { return typeof(this[name]); })")

            self.assertEqual("Boolean", typename("var_bool"))
            self.assertEqual("Number", typename("var_int"))
            self.assertEqual("Number", typename("var_float"))
            self.assertEqual("String", typename("var_str"))
            self.assertEqual("Date", typename("var_datetime"))
            self.assertEqual("Date", typename("var_date"))
            self.assertEqual("Date", typename("var_time"))

            self.assertEqual("MyInteger", typename("var_myint"))
            self.assertEqual("MyString", typename("var_mystr"))
            self.assertEqual("MyDateTime", typename("var_mytime"))

            self.assertEqual("function", typeof("var_myint"))
            self.assertEqual("function", typeof("var_mystr"))
            self.assertEqual("function", typeof("var_mytime"))

    def testJavascriptWrapper(self):
        with STPyV8.JSContext() as ctxt:
            self.assertEqual(type(None), type(ctxt.eval("null")))
            self.assertEqual(type(None), type(ctxt.eval("undefined")))
            self.assertEqual(bool, type(ctxt.eval("true")))
            self.assertEqual(str, type(ctxt.eval("'test'")))
            self.assertEqual(int, type(ctxt.eval("123")))
            self.assertEqual(float, type(ctxt.eval("3.14")))
            self.assertEqual(datetime.datetime, type(ctxt.eval("new Date()")))
            self.assertEqual(STPyV8.JSArray, type(ctxt.eval("[1, 2, 3]")))
            self.assertEqual(STPyV8.JSFunction, type(ctxt.eval("(function() {})")))
            self.assertEqual(STPyV8.JSObject, type(ctxt.eval("new Object()")))

    def testPythonWrapper(self):
        with STPyV8.JSContext() as ctxt:
            typeof = ctxt.eval("(function type(value) { return typeof value; })")
            protoof = ctxt.eval(
                "(function protoof(value) { return Object.prototype.toString.apply(value); })"
            )

            self.assertEqual("[object Null]", protoof(None))
            self.assertEqual("boolean", typeof(True))
            self.assertEqual("number", typeof(123))
            self.assertEqual("number", typeof(3.14))
            self.assertEqual("string", typeof("test"))

            self.assertEqual("[object Date]", protoof(datetime.datetime.now()))
            self.assertEqual("[object Date]", protoof(datetime.date.today()))
            self.assertEqual("[object Date]", protoof(datetime.time()))

            def test():
                pass

            self.assertEqual("[object Function]", protoof(abs))
            self.assertEqual("[object Function]", protoof(test))
            self.assertEqual("[object Function]", protoof(self.testPythonWrapper))
            self.assertEqual("[object Function]", protoof(int))

    def testFunction(self):
        with STPyV8.JSContext() as ctxt:
            func = ctxt.eval(
                """
                (function ()
                {
                    function a()
                    {
                        return "abc";
                    }

                    return a();
                })
                """
            )

            self.assertEqual("abc", str(func()))
            self.assertTrue(func is not None)
            self.assertFalse(func is None)

            func = ctxt.eval("(function test() {})")

            self.assertEqual("test", func.name)
            self.assertEqual("", func.resname)
            self.assertEqual(0, func.linenum)
            self.assertEqual(14, func.colnum)

            # FIXME
            # Why the setter doesn't work?
            #
            # func.name = "hello"
            # self.assertEqual("hello", func.name)

            func.setName("hello")
            self.assertEqual("hello", func.name)

    def testCall(self):
        class Hello:
            def __call__(self, name):
                return "hello " + name

        class Global(STPyV8.JSClass):
            hello = Hello()

        with STPyV8.JSContext(Global()) as ctxt:
            self.assertEqual("hello world", ctxt.eval("hello('world')"))

    def testJSFunction(self):
        with STPyV8.JSContext() as ctxt:
            hello = ctxt.eval("(function (name) { return 'Hello ' + name; })")

            self.assertTrue(isinstance(hello, STPyV8.JSFunction))
            self.assertEqual("Hello world", hello("world"))
            self.assertEqual("Hello world", hello.invoke(["world"]))

            obj = ctxt.eval(
                "({ 'name': 'world', 'hello': function (name) { return 'Hello ' + name + ' from ' + this.name; }})"
            )
            hello = obj.hello
            self.assertTrue(isinstance(hello, STPyV8.JSFunction))
            self.assertEqual("Hello world from world", hello("world"))

            tester = ctxt.eval("({ 'name': 'tester' })")
            self.assertEqual("Hello world from tester", hello.apply(tester, ["world"]))
            self.assertEqual(
                "Hello world from json", hello.apply({"name": "json"}, ["world"])
            )

    def testConstructor(self):
        with STPyV8.JSContext() as ctx:
            ctx.eval(
                """
                var Test = function() {
                    this.trySomething();
                };
                Test.prototype.trySomething = function() {
                    this.name = 'soirv8';
                };

                var Test2 = function(first_name, last_name) {
                    this.name = first_name + ' ' + last_name;
                };
                """
            )

            self.assertTrue(isinstance(ctx.locals.Test, STPyV8.JSFunction))

            test = STPyV8.JSObject.create(ctx.locals.Test)

            self.assertTrue(isinstance(ctx.locals.Test, STPyV8.JSObject))
            self.assertEqual("soirv8", test.name)

            test2 = STPyV8.JSObject.create(ctx.locals.Test2, ("John", "Doe"))

            self.assertEqual("John Doe", test2.name)

            test3 = STPyV8.JSObject.create(
                ctx.locals.Test2, ("John", "Doe"), {"email": "john.doe@randommail.com"}
            )

            self.assertEqual("john.doe@randommail.com", test3.email)

    def testJSError(self):
        with STPyV8.JSContext() as ctxt:
            try:
                ctxt.eval('throw "test"')
                self.fail()
            except Exception:  # pylint:disable=broad-except
                self.assertTrue(STPyV8.JSError, sys.exc_info()[0])

    def testErrorInfo(self):
        with STPyV8.JSContext():
            with STPyV8.JSEngine() as engine:
                try:
                    engine.compile(
                        """
                        function hello()
                        {
                            throw Error("hello world");
                        }

                        hello();""",
                        "test",
                        10,
                        10,
                    ).run()
                    self.fail()
                except STPyV8.JSError as e:
                    self.assertTrue(
                        "JSError: Error: hello world ( test @ 14 : 28 )  ->" in str(e)
                    )
                    self.assertEqual("Error", e.name)
                    self.assertEqual("hello world", e.message)
                    self.assertEqual("test", e.scriptName)
                    self.assertEqual(14, e.lineNum)
                    self.assertEqual(96, e.startPos)
                    self.assertEqual(97, e.endPos)
                    self.assertEqual(28, e.startCol)
                    self.assertEqual(29, e.endCol)
                    self.assertEqual(
                        'throw Error("hello world");', e.sourceLine.strip()
                    )

                    self.assertEqual(
                        "Error: hello world\n"
                        + "    at hello (test:14:35)\n"
                        + "    at test:17:25",
                        e.stackTrace,
                    )

    def testParseStack(self):
        self.assertEqual(
            [
                ("Error", "unknown source", None, None),
                ("test", "native", None, None),
                ("<anonymous>", "test0", 3, 5),
                ("f", "test1", 2, 19),
                ("g", "test2", 1, 15),
                (None, "test3", 1, None),
                (None, "test3", 1, 1),
            ],
            STPyV8.JSError.parse_stack(
                """Error: err
            at Error (unknown source)
            at test (native)
            at new <anonymous> (test0:3:5)
            at f (test1:2:19)
            at g (test2:1:15)
            at test3:1
            at test3:1:1"""
            ),
        )

    def testStackTrace(self):
        class Global(STPyV8.JSClass):
            def GetCurrentStackTrace(self, limit):  # pylint:disable=no-self-use
                return STPyV8.JSStackTrace.GetCurrentStackTrace(
                    limit, STPyV8.JSStackTrace.Options.Detailed
                )

        with STPyV8.JSContext(Global()) as ctxt:
            st = ctxt.eval(
                """
                function a()
                {
                    return GetCurrentStackTrace(4);
                }
                function b()
                {
                    return eval("a()");
                }
                function c()
                {
                    return new b();
                }
            c();""",
                "test",
            )

            self.assertEqual(4, len(st))
            self.assertEqual(
                "\tat a (test:4:28)\n\tat eval ((eval))\n\tat b (test:8:28)\n\tat c (test:12:28)\n",
                str(st),
            )
            self.assertEqual(
                "test.a (4:28)\n.eval (1:1) eval\ntest.b (8:28) constructor\ntest.c (12:28)",
                "\n".join(
                    [
                        "%s.%s (%d:%d)%s%s"
                        % (
                            f.scriptName,
                            f.funcName,
                            f.lineNum,
                            f.column,
                            " eval" if f.isEval else "",
                            " constructor" if f.isConstructor else "",
                        )
                        for f in st
                    ]
                ),
            )

    def testPythonException(self):
        class Global(STPyV8.JSClass):
            def raiseException(self):  # pylint:disable=no-self-use
                raise RuntimeError("Hello")

        with STPyV8.JSContext(Global()) as ctxt:
            ctxt.eval(
                """
                msg ="";
                try
                {
                    this.raiseException()
                }
                catch(e)
                {
                    msg += "catch " + e + ";";
                }
                finally
                {
                    msg += "finally";
                }"""
            )
            self.assertEqual("catch Error: Hello;finally", str(ctxt.locals.msg))

    def testExceptionMapping(self):
        class TestException(Exception):
            pass

        class Global(STPyV8.JSClass):
            def raiseIndexError(self):  # pylint:disable=no-self-use
                return [1, 2, 3][5]

            def raiseAttributeError(self):  # pylint:disable=no-self-use
                None.hello()

            def raiseSyntaxError(self):  # pylint:disable=no-self-use
                eval("???")  # pylint:disable=eval-used

            def raiseTypeError(self):  # pylint:disable=no-self-use
                int(sys)

            def raiseNotImplementedError(self):  # pylint:disable=no-self-use
                raise NotImplementedError("Not supported")

            def raiseExceptions(self):
                raise TestException()

        with STPyV8.JSContext(Global()) as ctxt:
            ctxt.eval("try { this.raiseIndexError(); } catch (e) { msg = e; }")

            self.assertEqual(
                "RangeError: list index out of range", str(ctxt.locals.msg)
            )

            ctxt.eval("try { this.raiseAttributeError(); } catch (e) { msg = e; }")

            self.assertEqual(
                "ReferenceError: 'NoneType' object has no attribute 'hello'",
                str(ctxt.locals.msg),
            )

            ctxt.eval("try { this.raiseSyntaxError(); } catch (e) { msg = e; }")

            self.assertEqual("SyntaxError: invalid syntax", str(ctxt.locals.msg))

            ctxt.eval("try { this.raiseTypeError(); } catch (e) { msg = e; }")

            if sys.version_info.major >= 3:
                if sys.version_info.minor > 9:
                    self.assertEqual(
                        "TypeError: int() argument must be a string, a bytes-like object or a real number, not 'module'",
                        str(ctxt.locals.msg),
                    )
                else:
                    self.assertEqual(
                        "TypeError: int() argument must be a string, a bytes-like object or a number, not 'module'",
                        str(ctxt.locals.msg),
                    )

            ctxt.eval("try { this.raiseNotImplementedError(); } catch (e) { msg = e; }")

            self.assertEqual("Error: Not supported", str(ctxt.locals.msg))

            self.assertRaises(TestException, ctxt.eval, "this.raiseExceptions();")

    def testArray(self):
        with STPyV8.JSContext() as ctxt:
            array = ctxt.eval(
                """
                var array = new Array();

                for (i=0; i<10; i++)
                {
                    array[i] = 10-i;
                }

                array;
                """
            )

            self.assertTrue(isinstance(array, STPyV8.JSArray))
            self.assertEqual(10, len(array))

            self.assertTrue(5 in array)
            self.assertFalse(15 in array)

            self.assertEqual(10, len(array))

            for i in range(10):
                self.assertEqual(10 - i, array[i])

            array[5] = 0

            self.assertEqual(0, array[5])

            del array[5]

            self.assertEqual(None, array[5])

            # array         [10, 9, 8, 7, 6, None, 4, 3, 2, 1]
            # array[4:7]                  4^^^^^^^^^7
            # array[-3:-1]                         -3^^^^^^-1
            # array[0:0]    []

            self.assertEqual([6, None, 4], array[4:7])
            self.assertEqual([3, 2], array[-3:-1])
            self.assertEqual([], array[0:0])

            self.assertEqual([10, 9, 8, 7, 6, None, 4, 3, 2, 1], list(array))

            array[1:3] = [9, 9]

            self.assertEqual([10, 9, 9, 7, 6, None, 4, 3, 2, 1], list(array))

            array[5:7] = [8, 8]

            self.assertEqual([10, 9, 9, 7, 6, 8, 8, 3, 2, 1], list(array))

            del array[1]

            self.assertEqual([10, None, 9, 7, 6, 8, 8, 3, 2, 1], list(array))

            ctxt.locals.array1 = STPyV8.JSArray(5)
            ctxt.locals.array2 = STPyV8.JSArray([1, 2, 3, 4, 5])

            for i in range(len(ctxt.locals.array2)):
                ctxt.locals.array1[i] = ctxt.locals.array2[i] * 10

            ctxt.eval(
                """
                var sum = 0;

                for (i=0; i<array1.length; i++)
                    sum += array1[i]

                for (i=0; i<array2.length; i++)
                    sum += array2[i]
                """
            )

            self.assertEqual(165, ctxt.locals.sum)

            ctxt.locals.array3 = [1, 2, 3, 4, 5]
            self.assertTrue(ctxt.eval("array3[1] === 2"))
            self.assertTrue(ctxt.eval("array3[9] === undefined"))

            args = [
                [
                    "a = Array(7); for(i=0; i<a.length; i++) a[i] = i; a[3] = undefined; a[a.length-1]; a",
                    "0,1,2,,4,5,6",
                    [0, 1, 2, None, 4, 5, 6],
                ],
                [
                    "a = Array(7); for(i=0; i<a.length - 1; i++) a[i] = i; a[a.length-1]; a",
                    "0,1,2,3,4,5,",
                    [0, 1, 2, 3, 4, 5, None],
                ],
                [
                    "a = Array(7); for(i=1; i<a.length; i++) a[i] = i; a[a.length-1]; a",
                    ",1,2,3,4,5,6",
                    [None, 1, 2, 3, 4, 5, 6],
                ],
            ]

            for arg in args:
                array = ctxt.eval(arg[0])

                self.assertEqual(arg[1], str(array))
                self.assertEqual(arg[2], [array[i] for i in range(len(array))])

            self.assertEqual(
                3,
                ctxt.eval("(function (arr) { return arr.length; })")(
                    STPyV8.JSArray([1, 2, 3])
                ),
            )
            self.assertEqual(
                2,
                ctxt.eval("(function (arr, idx) { return arr[idx]; })")(
                    STPyV8.JSArray([1, 2, 3]), 1
                ),
            )
            self.assertEqual(
                "[object Array]",
                ctxt.eval(
                    "(function (arr) { return Object.prototype.toString.call(arr); })"
                )(STPyV8.JSArray([1, 2, 3])),
            )
            self.assertEqual(
                "[object Array]",
                ctxt.eval(
                    "(function (arr) { return Object.prototype.toString.call(arr); })"
                )(STPyV8.JSArray((1, 2, 3))),
            )
            self.assertEqual(
                "[object Array]",
                ctxt.eval(
                    "(function (arr) { return Object.prototype.toString.call(arr); })"
                )(STPyV8.JSArray(list(range(3)))),
            )

    def testArraySlices(self):
        with STPyV8.JSContext() as ctxt:
            array = ctxt.eval(
                """
                var array = new Array();
                array;
            """
            )

            array[2:4] = [42, 24]
            # array         [None, None, 42, 24]
            self.assertEqual(len(array), 4)
            self.assertEqual(ctxt.eval("array[0]"), None)
            self.assertEqual(ctxt.eval("array[1]"), None)
            self.assertEqual(ctxt.eval("array[2]"), 42)
            self.assertEqual(ctxt.eval("array[3]"), 24)

            array[2:4] = [1, 2]
            # array         [None, None, 1, 2]
            self.assertEqual(len(array), 4)
            self.assertEqual(ctxt.eval("array[0]"), None)
            self.assertEqual(ctxt.eval("array[1]"), None)
            self.assertEqual(ctxt.eval("array[2]"), 1)
            self.assertEqual(ctxt.eval("array[3]"), 2)

            array[0:4:2] = [7, 8]
            # array         [7, None, 8, 2]
            self.assertEqual(len(array), 4)
            self.assertEqual(ctxt.eval("array[0]"), 7)
            self.assertEqual(ctxt.eval("array[1]"), None)
            self.assertEqual(ctxt.eval("array[2]"), 8)
            self.assertEqual(ctxt.eval("array[3]"), 2)

            array[1:4] = [10]
            # array         [7, 10, None, None]
            self.assertEqual(len(array), 4)
            self.assertEqual(ctxt.eval("array[0]"), 7)
            self.assertEqual(ctxt.eval("array[1]"), 10)
            self.assertEqual(ctxt.eval("array[2]"), None)
            self.assertEqual(ctxt.eval("array[3]"), None)

            array[0:7] = [0, 1, 2]
            # array         [0, 1, 2, None, None, None, None]
            self.assertEqual(len(array), 7)
            self.assertEqual(ctxt.eval("array[0]"), 0)
            self.assertEqual(ctxt.eval("array[1]"), 1)
            self.assertEqual(ctxt.eval("array[2]"), 2)
            self.assertEqual(ctxt.eval("array[3]"), None)
            self.assertEqual(ctxt.eval("array[4]"), None)
            self.assertEqual(ctxt.eval("array[5]"), None)
            self.assertEqual(ctxt.eval("array[6]"), None)

            array[0:7] = [0, 1, 2, 3, 4, 5, 6]
            # array         [0, 1, 2, 3, 4, 5, 6]
            self.assertEqual(len(array), 7)
            self.assertEqual(ctxt.eval("array[0]"), 0)
            self.assertEqual(ctxt.eval("array[1]"), 1)
            self.assertEqual(ctxt.eval("array[2]"), 2)
            self.assertEqual(ctxt.eval("array[3]"), 3)
            self.assertEqual(ctxt.eval("array[4]"), 4)
            self.assertEqual(ctxt.eval("array[5]"), 5)
            self.assertEqual(ctxt.eval("array[6]"), 6)

            del array[0:2]
            # array         [None, None, 2, 3, 4, 5, 6]
            self.assertEqual(len(array), 7)
            self.assertEqual(ctxt.eval("array[0]"), None)
            self.assertEqual(ctxt.eval("array[1]"), None)
            self.assertEqual(ctxt.eval("array[2]"), 2)
            self.assertEqual(ctxt.eval("array[3]"), 3)
            self.assertEqual(ctxt.eval("array[4]"), 4)
            self.assertEqual(ctxt.eval("array[5]"), 5)
            self.assertEqual(ctxt.eval("array[6]"), 6)

            del array[3:7:2]
            # array         [None, None, 2, None, 4, None, 6]
            self.assertEqual(len(array), 7)
            self.assertEqual(ctxt.eval("array[0]"), None)
            self.assertEqual(ctxt.eval("array[1]"), None)
            self.assertEqual(ctxt.eval("array[2]"), 2)
            self.assertEqual(ctxt.eval("array[3]"), None)
            self.assertEqual(ctxt.eval("array[4]"), 4)
            self.assertEqual(ctxt.eval("array[5]"), None)
            self.assertEqual(ctxt.eval("array[6]"), 6)

    def testMultiDimArray(self):
        with STPyV8.JSContext() as ctxt:
            ret = ctxt.eval(
                """
                ({
                    'test': function(){
                        return  [
                            [ 1, 'abla' ],
                            [ 2, 'ajkss' ],
                        ]
                    }
                })
                """
            ).test()

            self.assertEqual([[1, "abla"], [2, "ajkss"]], convert(ret))

    def testLazyConstructor(self):
        class Globals(STPyV8.JSClass):
            def __init__(self):
                self.array = STPyV8.JSArray([1, 2, 3])

        with STPyV8.JSContext(Globals()) as ctxt:
            self.assertEqual(2, ctxt.eval("""array[1]"""))

    def testForEach(self):
        class NamedClass:
            foo = 1  # pylint:disable=disallowed-name

            def __init__(self):
                self.bar = 2  # pylint:disable=disallowed-name

            @property
            def foobar(self):
                return self.foo + self.bar

        def gen(x):
            for i in range(x):
                yield i

        with STPyV8.JSContext() as ctxt:
            func = ctxt.eval(
                """(function (k) {
                var result = [];
                for (var prop in k) {
                  result.push(prop);
                }
                return result;
            })"""
            )

            self.assertTrue(
                set(["bar", "foo", "foobar"]).issubset(set(func(NamedClass())))
            )
            self.assertEqual(["0", "1", "2"], list(func([1, 2, 3])))
            self.assertEqual(["0", "1", "2"], list(func((1, 2, 3))))
            self.assertEqual(["1", "2", "3"], list(func({"1": 1, "2": 2, "3": 3})))

            self.assertEqual(["0", "1", "2"], list(func(list(gen(3)))))

    def testDict(self):
        with STPyV8.JSContext() as ctxt:
            obj = ctxt.eval("var r = { 'a' : 1, 'b' : 2 }; r")

            self.assertEqual(1, obj.a)
            self.assertEqual(2, obj.b)

            self.assertEqual({"a": 1, "b": 2}, dict(obj))

            self.assertEqual(
                {
                    "a": 1,
                    "b": [1, 2, 3],
                    "c": {"str": "goofy", "float": 1.234, "obj": {"name": "john doe"}},
                    "d": True,
                    "e": None,
                },
                convert(
                    ctxt.eval(
                        """var x =
                             { a: 1,
                               b: [1, 2, 3],
                               c: { str: 'goofy',
                                    float: 1.234,
                                    obj: { name: 'john doe' }},
                               d: true,
                               e: null }; x"""
                    )
                ),
            )

    def testDate(self):
        with STPyV8.JSContext() as ctxt:
            now1 = ctxt.eval("new Date();")

            self.assertTrue(now1)

            now2 = datetime.datetime.now()

            delta = now2 - now1 if now2 > now1 else now1 - now2

            self.assertTrue(delta < datetime.timedelta(seconds=1))

            func = ctxt.eval("(function (d) { return d.toString(); })")

            now = datetime.datetime.now()

            self.assertTrue(
                str(func(now).startswith(now.strftime("%a %b %d %Y %H:%M:%S")))
            )

            ctxt.eval("function identity(x) { return x; }")

            now3 = now2.replace(microsecond=123000)
            self.assertEqual(now3, ctxt.locals.identity(now3))

    def testUnicode(self):
        with STPyV8.JSContext() as ctxt:
            self.assertEqual("人", ctxt.eval('"人"'))
            self.assertEqual("é", ctxt.eval('"é"'))

            func = ctxt.eval("(function (msg) { return msg.length; })")

            self.assertEqual(2, func("测试"))

    def testClassicStyleObject(self):
        class FileSystemWrapper:
            @property
            def cwd(self):
                return os.getcwd()

        class Global:
            @property
            def fs(self):
                return FileSystemWrapper()

        with STPyV8.JSContext(Global()) as ctxt:
            self.assertEqual(os.getcwd(), ctxt.eval("fs.cwd"))

    def testRefCount(self):
        count = sys.getrefcount(None)

        class Global(STPyV8.JSClass):
            pass

        g = Global()
        g_refs = sys.getrefcount(g)

        with STPyV8.JSContext(g) as ctxt:
            ctxt.eval(
                """
                var none = null;
            """
            )

            self.assertTrue(count <= sys.getrefcount(None) <= count + 2)

            ctxt.eval(
                """
                var none = null;
            """
            )

            self.assertTrue(count <= sys.getrefcount(None) <= count + 2)

            del ctxt

        self.assertEqual(g_refs, sys.getrefcount(g))

    def testProperty(self):
        class Global(STPyV8.JSClass):
            def __init__(self, name):
                self._name = name

            def getname(self):
                return self._name

            def setname(self, name):
                self._name = name

            def delname(self):
                self._name = "deleted"

            name = property(getname, setname, delname)

        g = Global("world")

        with STPyV8.JSContext(g) as ctxt:
            self.assertEqual("world", ctxt.eval("name"))
            self.assertEqual("foobar", ctxt.eval("this.name = 'foobar';"))
            self.assertEqual("foobar", ctxt.eval("name"))
            self.assertTrue(ctxt.eval("delete name"))

        with STPyV8.JSContext() as ctxt:
            self.assertEqual("world", ctxt.eval("name = 'world';"))
            self.assertTrue(ctxt.eval("delete name;"))
            self.assertRaises(ReferenceError, ctxt.eval, "name")

            ctxt.eval(
                """
                let obj = {};
                obj.__defineGetter__('name', function() {
                    return 'test';
                });
            """
            )
            self.assertEqual("test", ctxt.eval("obj.name"))

    def testGetterAndSetter(self):
        class Global(STPyV8.JSClass):
            def __init__(self, testval):
                self.testval = testval

        with STPyV8.JSContext(Global("Test Value A")) as ctxt:
            self.assertEqual("Test Value A", ctxt.locals.testval)
            ctxt.eval(
                """
               this.__defineGetter__("test", function() {
                   return this.testval;
               });
               this.__defineSetter__("test", function(val) {
                   this.testval = val;
               });
           """
            )
            self.assertEqual("Test Value A", ctxt.locals.test)

            ctxt.eval("test = 'Test Value B';")

            self.assertEqual("Test Value B", ctxt.locals.test)

    def testReferenceCount(self):
        class Hello:
            def say(self):
                pass

            def __del__(self):
                owner.deleted = True  # pylint:disable=undefined-variable

        def test():
            with STPyV8.JSContext() as ctxt:
                fn = ctxt.eval("(function (obj) { obj.say(); })")

                obj = Hello()

                self.assertEqual(2, sys.getrefcount(obj))

                fn(obj)

                self.assertEqual(4, sys.getrefcount(obj))

                del obj

        test()

    def testNullInString(self):
        with STPyV8.JSContext() as ctxt:
            fn = ctxt.eval("(function (s) { return s; })")

            self.assertEqual("hello \0 world", fn("hello \0 world"))

    def testLivingObjectCache(self):
        class Global(STPyV8.JSClass):
            i = 1
            b = True
            o = object()

        with STPyV8.JSContext(Global()) as ctxt:
            self.assertTrue(ctxt.eval("i == i"))
            self.assertTrue(ctxt.eval("b == b"))
            self.assertTrue(ctxt.eval("o == o"))

    @pytest.mark.skipif(STPYV8_DEBUG, reason="Not a test for debug mode")
    def testMemoryLeak(self):
        with STPyV8.JSIsolate():
            with STPyV8.JSContext() as ctxt:
                inner = ctxt.eval("i => Array(1<<20).fill(i)")
                outer = ctxt.eval("o => o[0]")

                for i in range(1000):
                    outer(inner(i))

                self.assertEqual(999, i)

    def testNamedSetter(self):
        class Obj(STPyV8.JSClass):
            @property
            def p(self):
                return self._p

            @p.setter
            def p(self, value):
                self._p = value

        class Global(STPyV8.JSClass):
            def __init__(self):
                self.obj = Obj()
                self.d = {}
                self.p = None

        with STPyV8.JSContext(Global()) as ctxt:
            ctxt.eval(
                """
            x = obj;
            x.y = 10;
            x.p = 10;
            d.y = 10;
            """
            )
            self.assertEqual(10, ctxt.eval("obj.y"))
            self.assertEqual(10, ctxt.eval("obj.p"))
            self.assertEqual(10, ctxt.locals.d["y"])

    def testWatch(self):
        class Obj(STPyV8.JSClass):
            def __init__(self):
                self.p = 1

        class Global(STPyV8.JSClass):
            def __init__(self):
                self.o = Obj()

        with STPyV8.JSContext(Global()) as ctxt:
            ctxt.eval(
                """
            o.watch("p", function (id, oldval, newval) {
                return oldval + newval;
            });
            """
            )

            self.assertEqual(1, ctxt.eval("o.p"))

            ctxt.eval("o.p = 2;")

            self.assertEqual(3, ctxt.eval("o.p"))

            ctxt.eval("delete o.p;")

            self.assertEqual(None, ctxt.eval("o.p"))

            ctxt.eval("o.p = 2;")

            self.assertEqual(2, ctxt.eval("o.p"))

            ctxt.eval("o.unwatch('p');")

            ctxt.eval("o.p = 1;")

            self.assertEqual(1, ctxt.eval("o.p"))

    def testReferenceError(self):
        class Global(STPyV8.JSClass):
            def __init__(self):
                self.s = self

        with STPyV8.JSContext(Global()) as ctxt:
            self.assertRaises(ReferenceError, ctxt.eval, "x")

            self.assertTrue(ctxt.eval("typeof(x) === 'undefined'"))

            self.assertTrue(ctxt.eval("typeof(String) === 'function'"))

            self.assertTrue(ctxt.eval("typeof(s.String) === 'undefined'"))

            self.assertTrue(ctxt.eval("typeof(s.z) === 'undefined'"))

    def testRaiseExceptionInGetter(self):
        class Document(STPyV8.JSClass):
            def __getattr__(self, name):
                if name == "y":
                    raise TypeError()

                return STPyV8.JSClass.__getattr__(self, name)

        class Global(STPyV8.JSClass):
            def __init__(self):
                self.document = Document()

        with STPyV8.JSContext(Global()) as ctxt:
            self.assertEqual(None, ctxt.eval("document.x"))
            self.assertRaises(TypeError, ctxt.eval, "document.y")

    def testUndefined(self):
        class Global(STPyV8.JSClass):
            def returnNull(self):  # pylint:disable=no-self-use
                return STPyV8.JSNull()

            def returnUndefined(self):  # pylint:disable=no-self-use
                return STPyV8.JSUndefined()

            def returnNone(self):  # pylint:disable=no-self-use
                return None

        with STPyV8.JSContext(Global()) as ctxt:
            self.assertFalse(bool(STPyV8.JSNull()))
            self.assertFalse(bool(STPyV8.JSUndefined()))

            self.assertEqual("null", str(STPyV8.JSNull()))
            self.assertEqual("undefined", str(STPyV8.JSUndefined()))

            self.assertTrue(ctxt.eval("null == returnNull()"))
            self.assertTrue(ctxt.eval("undefined == returnUndefined()"))
            self.assertTrue(ctxt.eval("null == returnNone()"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

import STPyV8


class TestEngine(unittest.TestCase):
    def testClassProperties(self):
        with STPyV8.JSContext():
            self.assertTrue(str(STPyV8.JSEngine.version).startswith("12."))
            self.assertFalse(STPyV8.JSEngine.dead)

    def testCompile(self):
        with STPyV8.JSContext():
            with STPyV8.JSEngine() as engine:
                s = engine.compile("1+2")

                self.assertTrue(isinstance(s, STPyV8.JSScript))

                self.assertEqual("1+2", s.source)
                self.assertEqual(3, int(s.run()))

                self.assertRaises(SyntaxError, engine.compile, "1+")

    def testUnicodeSource(self):
        class Global(STPyV8.JSClass):
            var = "测试"

            def __getattr__(self, name):
                if name:
                    return self.var

                return STPyV8.JSClass.__getattr__(self, name)

        g = Global()

        with STPyV8.JSContext(g) as ctxt:
            with STPyV8.JSEngine() as engine:
                src = """
                function 函数() { return 变量.length; }

                函数();

                var func = function () {};
                """

                s = engine.compile(src)

                self.assertTrue(isinstance(s, STPyV8.JSScript))

                self.assertEqual(src, s.source)
                self.assertEqual(2, s.run())

                func_name = "函数"

                self.assertTrue(hasattr(ctxt.locals, func_name))

                func = getattr(ctxt.locals, func_name)

                self.assertTrue(isinstance(func, STPyV8.JSFunction))

                self.assertEqual(func_name, func.name)
                self.assertEqual("", func.resname)
                self.assertEqual(1, func.linenum)

                var_name = "变量"

                setattr(ctxt.locals, var_name, "测试长字符串")

                self.assertEqual(6, func())

                self.assertEqual("func", ctxt.locals.func.inferredname)

    def testEval(self):
        with STPyV8.JSContext() as ctxt:
            self.assertEqual(3, int(ctxt.eval("1+2")))

    def testGlobal(self):
        class Global(STPyV8.JSClass):
            version = "1.0"

        with STPyV8.JSContext(Global()) as ctxt:
            _vars = ctxt.locals

            # getter
            self.assertEqual(Global.version, str(_vars.version))
            self.assertEqual(Global.version, str(ctxt.eval("version")))

            self.assertRaises(ReferenceError, ctxt.eval, "nonexists")

            # setter
            self.assertEqual(2.0, float(ctxt.eval("version = 2.0")))

            self.assertEqual(2.0, float(_vars.version))

    def testThis(self):
        class Global(STPyV8.JSClass):
            version = 1.0

        with STPyV8.JSContext(Global()) as ctxt:
            self.assertEqual("[object Global]", str(ctxt.eval("this")))
            self.assertEqual(1.0, float(ctxt.eval("this.version")))

    def testObjectBuiltInMethods(self):
        class Global(STPyV8.JSClass):
            version = 1.0

        with STPyV8.JSContext(Global()) as ctxt:
            self.assertEqual("[object Global]", str(ctxt.eval("this.toString()")))
            self.assertEqual("[object Global]", str(ctxt.eval("this.toLocaleString()")))
            self.assertEqual(Global.version, float(ctxt.eval("this.valueOf()").version))
            self.assertTrue(bool(ctxt.eval('this.hasOwnProperty("version")')))
            self.assertFalse(ctxt.eval('this.hasOwnProperty("nonexistent")'))

    def testPythonWrapper(self):
        class Global(STPyV8.JSClass):
            s = [1, 2, 3]
            d = {"a": {"b": "c"}, "d": ["e", "f"]}

        g = Global()

        with STPyV8.JSContext(g) as ctxt:
            ctxt.eval(
                """
                s[2] = s[1] + 2;
                s[0] = s[1];
                delete s[1];
            """
            )
            self.assertEqual([2, 4], g.s)
            self.assertEqual("c", ctxt.eval("d.a.b"))
            self.assertEqual(["e", "f"], ctxt.eval("d.d"))
            ctxt.eval(
                """
                d.a.q = 4
                delete d.d
            """
            )
            self.assertEqual(4, g.d["a"]["q"])
            self.assertEqual(None, ctxt.eval("d.d"))

    def _testMemoryAllocationCallback(self):
        alloc = {}

        def callback(space, action, size):
            alloc[(space, action)] = alloc.setdefault((space, action), 0) + size

        STPyV8.JSEngine.setMemoryAllocationCallback(callback)

        with STPyV8.JSContext() as ctxt:
            self.assertFalse(
                (STPyV8.JSObjectSpace.Code, STPyV8.JSAllocationAction.alloc) in alloc
            )

            ctxt.eval("var o = new Array(1000);")

            self.assertTrue(
                (STPyV8.JSObjectSpace.Code, STPyV8.JSAllocationAction.alloc) in alloc
            )

        STPyV8.JSEngine.setMemoryAllocationCallback(None)

    def _testOutOfMemory(self):
        with STPyV8.JSIsolate():
            STPyV8.JSEngine.setMemoryLimit(
                max_young_space_size=16 * 1024, max_old_space_size=4 * 1024 * 1024
            )

            with STPyV8.JSContext() as ctxt:
                STPyV8.JSEngine.ignoreOutOfMemoryException()

                ctxt.eval("var a = new Array(); while(true) a.push(a);")

                self.assertTrue(ctxt.hasOutOfMemoryException)

                STPyV8.JSEngine.setMemoryLimit()

                STPyV8.JSEngine.collect()

    def testStackLimit(self):
        with STPyV8.JSIsolate():
            with STPyV8.JSContext() as ctxt:
                STPyV8.JSEngine.setStackLimit(256 * 1024)
                oldStackSize = ctxt.eval(
                    "var maxStackSize = function(i){try{(function m(){++i&&m()}())}catch(e){return i}}(0); maxStackSize"
                )

        with STPyV8.JSIsolate():
            with STPyV8.JSContext() as ctxt:
                STPyV8.JSEngine.setStackLimit(512 * 1024)
                newStackSize = ctxt.eval(
                    "var maxStackSize = function(i){try{(function m(){++i&&m()}())}catch(e){return i}}(0); maxStackSize"
                )

        self.assertTrue(newStackSize > oldStackSize * 2)



#!/usr/bin/env python

import logging
import shutil
import subprocess

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.install import install
from wheel.bdist_wheel import bdist_wheel

from settings import *  # pylint:disable=wildcard-import,unused-wildcard-import

log = logging.getLogger()

ICU_DATA_PACKAGE_FOLDER = os.path.join(os.pardir, os.pardir, "stpyv8-icu")
ICU_DATA_V8_FILE_PATH = os.path.join("v8", "out.gn", "x64.release.sample", "icudtl.dat")


def exec_cmd(cmdline, *args, **kwargs):
    msg = kwargs.get("msg")
    cwd = kwargs.get("cwd", ".")
    output = kwargs.get("output")

    if msg:
        print(msg)

    cmdline = " ".join([cmdline] + list(args))

    proc = subprocess.Popen(
        cmdline,  # pylint:disable=consider-using-with
        shell=kwargs.get("shell", True),
        cwd=cwd,
        env=kwargs.get("env"),
        stdout=subprocess.PIPE if output else None,
        stderr=subprocess.PIPE if output else None,
    )

    stdout, stderr = proc.communicate()

    succeeded = proc.returncode == 0

    if not succeeded:
        log.error("%s failed: code = %d", msg or "Execute command", proc.returncode)

        if output:
            log.debug(stderr)

    return succeeded, stdout, stderr if output else succeeded


def install_depot():
    if not os.path.exists(DEPOT_HOME):
        exec_cmd(
            "git clone --depth 1",
            DEPOT_GIT_URL,
            DEPOT_HOME,
            cwd=os.path.dirname(DEPOT_HOME),
            msg="Cloning depot tools",
        )

        return

    # depot_tools updates itself automatically when running gclient tool
    if os.path.isfile(os.path.join(DEPOT_HOME, "gclient")):
        success, stdout, __ = exec_cmd(
            os.path.join(DEPOT_HOME, "gclient"),  # pylint:disable=unused-variable
            "--version",
            cwd=DEPOT_HOME,
            output=True,
            msg="Found depot tools",
        )

        if not success:
            exit(1)


def checkout_v8():
    install_depot()

    if not os.path.exists(V8_HOME):
        success, _, __ = exec_cmd(
            os.path.join(DEPOT_HOME, "fetch"),
            "--no-history",
            "v8",
            cwd=os.path.dirname(V8_HOME),
            msg="Fetching Google V8 code",
        )

        if not success:
            exit(1)

    success, _, __ = exec_cmd(
        "git fetch --tags --quiet",
        cwd=V8_HOME,
        msg="Fetching the release tag information",
    )

    if not success:
        exit(1)

    success, _, __ = exec_cmd(
        "git checkout", V8_GIT_TAG, cwd=V8_HOME, msg=f"Checkout Google V8 v{V8_GIT_TAG}"
    )

    if not success:
        exit(1)

    success, _, __ = exec_cmd(
        os.path.join(DEPOT_HOME, "gclient"),
        "sync",
        "-D",
        "-j8",
        cwd=os.path.dirname(V8_HOME),
        msg="Syncing Google V8 code",
    )

    if not success:
        exit(1)

    # On Linux, install additional dependencies, per
    # https://v8.dev/docs/build step 4
    if (
        sys.platform
        in (
            "linux",
            "linux2",
        )
        and v8_deps_linux
    ):
        success, _, __ = exec_cmd(
            "./v8/build/install-build-deps.sh",
            cwd=os.path.dirname(V8_HOME),
            msg="Installing additional linux dependencies",
        )

        if not success:
            exit(1)


def build_v8():
    args = f"gen {os.path.join('out.gn', 'x64.release.sample')} --args=\"{GN_ARGS}\""
    success, _, __ = exec_cmd(
        os.path.join(DEPOT_HOME, "gn"),
        args,
        cwd=V8_HOME,
        msg=f"Generate build scripts for V8 (v{V8_GIT_TAG})",
    )

    if not success:
        exit(1)

    success, _, __ = exec_cmd(
        os.path.join(DEPOT_HOME, "ninja"),
        f"-C {os.path.join('out.gn', 'x64.release.sample')} v8_monolith",
        cwd=V8_HOME,
        msg="Build V8 with ninja",
    )

    if not success:
        exit(1)


def clean_stpyv8():
    build_folder = os.path.join(STPYV8_HOME, "build")

    if os.path.exists(os.path.join(build_folder)):
        shutil.rmtree(build_folder)


def prepare_v8():
    try:
        checkout_v8()
        build_v8()
        # clean_stpyv8()
    except Exception as e:  # pylint:disable=broad-except
        log.error("Fail to checkout and build v8, %s", str(e))


class stpyv8_bdist_wheel(bdist_wheel):
    user_options = bdist_wheel.user_options + [
        ("skip-build-v8", None, "don't build v8")
    ]

    def initialize_options(self) -> None:
        bdist_wheel.initialize_options(self)

        self.skip_build_v8 = None

    def run(self):
        if not self.skip_build_v8:
            prepare_v8()

        bdist_wheel.run(self)


class stpyv8_build(build_ext):
    def run(self):
        V8_GIT_TAG = V8_GIT_TAG_STABLE  # pylint:disable=redefined-outer-name,unused-variable

        build_ext.run(self)


class stpyv8_develop(build_ext):
    def run(self):
        V8_GIT_TAG = V8_GIT_TAG_MASTER  # pylint:disable=redefined-outer-name,unused-variable
        prepare_v8()
        build_ext.run(self)


class stpyv8_install_v8(build_ext):
    def run(self):
        V8_GIT_TAG = V8_GIT_TAG_MASTER  # pylint:disable=redefined-outer-name,unused-variable
        prepare_v8()


class stpyv8_build_no_v8(build_ext):
    def run(self):
        clean_stpyv8()
        build_ext.run(self)


class stpyv8_install(install):
    def run(self):
        self.skip_build = True  # pylint:disable=attribute-defined-outside-init
        install.run(self)


class stpyv8_checkout_v8(build_ext):
    def run(self):
        checkout_v8()


stpyv8 = Extension(
    name="_STPyV8",
    sources=[os.path.join("src", source) for source in source_files],
    define_macros=macros,
    include_dirs=include_dirs,
    library_dirs=library_dirs,
    libraries=libraries,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
)

setup(
    name="stpyv8",
    version=STPYV8_VERSION,
    description="Python Wrapper for Google V8 Engine",
    long_description=open("README.md").read(),
    platforms="x86",
    author="Philip Syme, Angelo Dell'Aera",
    url="https://github.com/cloudflare/stpyv8",
    license="Apache License 2.0",
    py_modules=["STPyV8"],
    ext_modules=[stpyv8],
    install_requires=["wheel", "importlib_resources; python_version < '3.10'"],
    data_files=[
        (ICU_DATA_PACKAGE_FOLDER, [ICU_DATA_V8_FILE_PATH]),
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: C++",
        "Programming Language :: Python",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    long_description_content_type="text/markdown",
    cmdclass=dict(
        bdist_wheel=stpyv8_bdist_wheel,
        build_ext=stpyv8_build,
        develop=stpyv8_develop,
        v8=stpyv8_install_v8,
        stpyv8=stpyv8_build_no_v8,
        install=stpyv8_install,
        checkout_v8=stpyv8_checkout_v8,
    ),
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
import threading
import unittest
import pytest

import STPyV8


class TestThread(unittest.TestCase):
    @pytest.mark.order(1)
    def testMultiPythonThread(self):
        class Global:
            count = 0
            started = threading.Event()
            finished = threading.Semaphore(0)

            def sleep(self, ms):
                time.sleep(ms / 1000.0)

                self.count += 1

        g = Global()

        def run():
            with STPyV8.JSIsolate():
                with STPyV8.JSContext(g) as ctxt:
                    ctxt.eval(
                        """
                        started.wait();

                        for (i=0; i<10; i++)
                        {
                            sleep(100);
                        }

                        finished.release();
                    """
                    )

        t = threading.Thread(target=run)
        t.start()

        now = time.time()

        self.assertEqual(0, g.count)

        g.started.set()
        g.finished.acquire()

        self.assertEqual(10, g.count)

        self.assertTrue((time.time() - now) >= 1)
        t.join()

    @pytest.mark.order(2)
    def testMultiJavascriptThread(self):
        class Global(STPyV8.JSContext):
            result = []

            def add(self, value):
                with STPyV8.JSUnlocker():
                    self.result.append(value)

        g = Global()

        def run():
            with STPyV8.JSIsolate():
                with STPyV8.JSContext(g) as ctxt:
                    ctxt.eval(
                        """
                        for (i=0; i<10; i++)
                            add(i);
                    """
                    )

        threads = [threading.Thread(target=run), threading.Thread(target=run)]

        with STPyV8.JSLocker():
            for t in threads:
                t.start()

        for t in threads:
            t.join()

        self.assertEqual(20, len(g.result))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

import STPyV8


class TestLocker(unittest.TestCase):
    def testLocker(self):
        with STPyV8.JSIsolate():
            self.assertFalse(STPyV8.JSLocker.locked)

            with STPyV8.JSLocker() as outter_locker:
                self.assertTrue(STPyV8.JSLocker.locked)

                self.assertTrue(outter_locker)

                with STPyV8.JSLocker() as inner_locker:
                    self.assertTrue(STPyV8.JSLocker.locked)

                    self.assertTrue(outter_locker)
                    self.assertTrue(inner_locker)

                    with STPyV8.JSUnlocker():
                        self.assertFalse(STPyV8.JSLocker.locked)

                        self.assertTrue(outter_locker)
                        self.assertTrue(inner_locker)

                    self.assertTrue(STPyV8.JSLocker.locked)

            self.assertFalse(STPyV8.JSLocker.locked)

            locker = STPyV8.JSLocker()

        with STPyV8.JSContext():
            self.assertRaises(RuntimeError, locker.__enter__)
            self.assertRaises(RuntimeError, locker.__exit__, None, None, None)

        del locker



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import unittest

ICU_DATA_FOLDERS_UNIX = (
    "/usr/share/stpyv8",
    os.path.expanduser("~/.local/share/stpyv8"),
)
ICU_DATA_FOLDERS_OSX = (
    "/Library/Application Support/STPyV8",
    os.path.expanduser("~/Library/Application Support/STPyV8"),
)
ICU_DATA_FOLDERS_WINDOWS = (
    os.path.join(os.environ["PROGRAMDATA"], "STPyV8")
    if "PROGRAMDATA" in os.environ
    else None,
    os.path.join(os.environ["APPDATA"], "STPyV8") if "APPDATA" in os.environ else None,
)

icu_data_folders = None
if os.name in ("posix",):
    icu_data_folders = (
        ICU_DATA_FOLDERS_OSX if sys.platform in ("darwin",) else ICU_DATA_FOLDERS_UNIX
    )
else:
    icu_data_folders = ICU_DATA_FOLDERS_WINDOWS


class TestICU(unittest.TestCase):
    def testIcu(self):
        icu_paths = (os.path.join(folder, "icudtl.dat") for folder in icu_data_folders)
        self.assertTrue(any(os.path.exists(icu) for icu in icu_paths))
